from pydantic import BaseModel
from typing import List, Union
import yaml
import re
from typing import Optional
from flo_ai.models.exception import FloValidationException


KIND_SUPERVISED_TEAM = 'FloRoutedTeam'
KIND_FLO_AGENT = 'FloAgent'

yaml_kinds = [KIND_SUPERVISED_TEAM, KIND_FLO_AGENT]


class KeyValueArgs(BaseModel):
    name: str
    value: str


class FilterArgs(BaseModel):
    name: str
    description: str
    type: str


class ToolConfig(BaseModel):
    name: str
    args: Optional[List[KeyValueArgs]] = None
    properties: Optional[List[KeyValueArgs]] = None
    filters: Optional[List[FilterArgs]] = None


class PromptStrategy(BaseModel):
    kind: str
    agent_name: str
    retries: int
    next: str | None = None


class MemberKey(BaseModel):
    name: str


class AgentConfig(BaseModel):
    name: str
    role: Optional[str] = None
    kind: Optional[str] = None
    job: Optional[str] = None
    tools: List[ToolConfig] = []
    to: Optional[List[MemberKey]] = None
    retry: Optional[int] = 1
    model: Optional[str] = None


class EdgeConfig(BaseModel):
    edge: List[str]
    type: Optional[str] = None
    rule: Optional[str] = None


class RouterConfig(BaseModel):
    name: str
    kind: str
    model: Optional[str] = None
    job: Optional[str] = None
    start_node: Optional[str] = None
    end_node: Union[Optional[str], List[str]] = None
    edges: Optional[List[EdgeConfig]] = None


class PlannerConfig(BaseModel):
    name: str


class TeamConfig(BaseModel):
    name: str
    kind: Optional[str] = None
    agents: Optional[List[AgentConfig]] = None
    subteams: Optional[List['TeamConfig']] = None
    router: Optional[RouterConfig] = None
    planner: Optional[PlannerConfig] = None


class FloRoutedTeamConfig(BaseModel):
    apiVersion: str
    kind: str
    name: str
    team: TeamConfig


class FloAgentConfig(BaseModel):
    apiVersion: str
    kind: Optional[str] = None
    name: str
    agent: AgentConfig


def to_supervised_team(yaml_str: str) -> FloRoutedTeamConfig:
    parsed_data = yaml.safe_load(yaml_str)
    kind = parsed_data['kind']
    if kind == KIND_SUPERVISED_TEAM:
        flo_supervised_team = FloRoutedTeamConfig(**parsed_data)
        validate_sup_team_config(flo_supervised_team)
        return flo_supervised_team
    elif kind == KIND_FLO_AGENT:
        flo_agent = FloAgentConfig(**parsed_data)
        validate_sup_team_config(flo_agent)
        return flo_agent
    else:
        raise FloValidationException('Unknown kind: {}'.format(kind))


def validate_sup_team_config(flo: FloRoutedTeamConfig):
    if flo.kind == KIND_FLO_AGENT:
        return
    if flo.name is None or not is_valid_name(flo.name):
        raise FloValidationException(
            'Invalid agent name while creating the flow, expected: [^[a-z][a-z0-9_-]*$]'
        )


def is_valid_name(s: str) -> bool:
    pattern = r'^[a-z][a-z0-9_-]*$'
    return bool(re.match(pattern, s))



import logging
from typing import Any
from typing import Dict, Optional, Union
from logging.handlers import RotatingFileHandler
from dataclasses import dataclass

DEFAULT_LOGGER_NAME = 'FloAI'
DEFAULT_LOG_LEVEL = 'ERROR'

LEVEL_MAP = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL,
}


@dataclass
class FloLogConfig:
    name: str
    level: Union[str, int] = DEFAULT_LOG_LEVEL
    file_path: str = None
    max_bytes: int = 1048576

    def get_level(self) -> int:
        """Convert string level to logging level integer if needed"""
        if isinstance(self.level, str):
            return LEVEL_MAP.get(self.level.upper(), logging.ERROR)
        return self.level


class FloLoggerUtil(logging.Logger):
    def __init__(self, config: FloLogConfig):
        level = config.get_level()
        super().__init__(config.name, level)
        self.setLevel(level)
        for handler in self.handlers:
            self.removeHandler(handler)
        self.setConfig(config)

    def setConfig(self, config: FloLogConfig):
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(session)s - %(levelname)s - %(message)s'
        )

        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        console_handler.setLevel(self.level)
        self.addHandler(console_handler)

        if config.file_path:
            file_handler = RotatingFileHandler(
                config.file_path, maxBytes=config.max_bytes
            )
            file_handler.setFormatter(formatter)
            file_handler.setLevel(self.level)
            self.addHandler(file_handler)

    def setLevel(self, level: Union[str, int]) -> None:
        if isinstance(level, str):
            level = LEVEL_MAP.get(level.upper(), logging.ERROR)
        super().setLevel(level)
        for handler in self.handlers:
            print('Setting levels in handler: ' + str(level))
            handler.setLevel(level)

    def _log(
        self, level: int, msg: str, session: Optional[str] = None, *args, **kwargs
    ):
        if not self.isEnabledFor(level):
            return
        if kwargs.get('extra') is None:
            kwargs['extra'] = {}
        kwargs['extra']['session'] = f'[{session.session_id}]' if session else '[-]'
        super()._log(level, msg, args, **kwargs)

    def debug(self, msg: str, session: Optional[Any] = None, *args, **kwargs):
        self._log(logging.DEBUG, msg, session, *args, **kwargs)

    def info(self, msg: str, session: Optional[Any] = None, *args, **kwargs):
        self._log(logging.INFO, msg, session, *args, **kwargs)

    def warning(self, msg: str, session: Optional[Any] = None, *args, **kwargs):
        self._log(logging.WARNING, msg, session, *args, **kwargs)

    def error(self, msg: str, session: Optional[Any] = None, *args, **kwargs):
        self._log(logging.ERROR, msg, session, *args, **kwargs)

    def critical(self, msg: str, session: Optional[Any] = None, *args, **kwargs):
        self._log(logging.CRITICAL, msg, session, *args, **kwargs)


logging_cache: Dict[str, FloLoggerUtil] = dict(
    {
        DEFAULT_LOGGER_NAME: FloLoggerUtil(
            FloLogConfig(DEFAULT_LOGGER_NAME, DEFAULT_LOG_LEVEL)
        )
    }
)


def get_logger(
    config: FloLogConfig = FloLogConfig(DEFAULT_LOGGER_NAME),
) -> FloLoggerUtil:
    if config.name not in logging_cache:
        logging_cache[config.name] = FloLoggerUtil(config)
    return logging_cache[config.name]


def set_log_level_internal(level: Union[str, int]) -> None:
    updated_logger = FloLoggerUtil(FloLogConfig(DEFAULT_LOGGER_NAME, level))
    logging_cache[DEFAULT_LOGGER_NAME] = updated_logger


def set_log_config_internal(config: FloLogConfig):
    updated_logger = FloLoggerUtil(config)
    logging_cache[DEFAULT_LOGGER_NAME] = updated_logger


def set_logger_internal(logger: logging.Logger):
    logging_cache[DEFAULT_LOGGER_NAME] = logger



from typing import Any, Union, Callable, Optional, Dict
from dataclasses import dataclass, field
from flo_ai.common.flo_logger import get_logger


@dataclass
class FloCallbackResponse:
    type: str
    name: Optional[str] = None
    model_name: Optional[str] = None
    input: Optional[str] = None
    output: Optional[str] = None
    error: Union[Exception, KeyboardInterrupt, None] = None
    args: Dict = field(default_factory=dict)


class FloToolCallback:
    def __init__(self) -> None:
        pass

    def on_tool_start(
        self, name: str, input: Any, **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        pass

    def on_tool_end(
        self, name: str, output: Any, **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        pass

    def on_tool_error(
        self, name: str, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        pass


class FloAgentCallback:
    def __init__(self) -> None:
        pass

    def on_agent_start(
        self, name: str, model_name: str, input: Any, **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        pass

    def on_agent_end(
        self, name: str, model_name: str, output: Any, **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        pass

    def on_agent_error(
        self,
        name: str,
        model_name: str,
        error: Union[Exception, KeyboardInterrupt],
        **kwargs: Any,
    ) -> Optional[FloCallbackResponse]:
        pass


class FloRouterCallback:
    def __init__(self) -> None:
        pass

    def on_router_start(
        self, name: str, model_name: str, input: Any, **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        pass

    def on_router_end(
        self, name: str, model_name: str, output: Any, **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        pass

    def on_router_error(
        self,
        name: str,
        model_name: str,
        error: Union[Exception, KeyboardInterrupt],
        **kwargs: Any,
    ) -> None:
        Optional[FloCallbackResponse]


def safe_call_cb(func, cb_response: FloCallbackResponse, ignore_error=True):
    try:
        func(cb_response)
    except Exception as e:
        if ignore_error:
            get_logger().warning(e)
        else:
            raise e


class FunctionalFloToolCallbackImpl(FloToolCallback):
    def __init__(self, func: Callable, ignore_error: bool = True) -> None:
        super().__init__()
        self.func = func
        self.ignore_error = ignore_error

    def on_tool_start(
        self, name: str, input: Any, **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        cb_response = FloRouterCallback('on_tool_start', name, input=input, args=kwargs)
        safe_call_cb(self.func, cb_response, self.ignore_error)
        return cb_response

    def on_tool_end(
        self, name: str, output: Any, **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        cb_response = FloRouterCallback('on_tool_end', name, output=output, args=kwargs)
        safe_call_cb(self.func, cb_response, self.ignore_error)
        return cb_response

    def on_tool_error(
        self, name: str, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Optional[FloCallbackResponse]:
        cb_response = FloRouterCallback('on_tool_error', name, error=error, args=kwargs)
        safe_call_cb(self.func, cb_response, self.ignore_error)
        return cb_response


class FunctionalFloAgentCallbackImpl(FloAgentCallback):
    def __init__(self, func: Callable, ignore_error: bool = True) -> None:
        super().__init__()
        self.func = func
        self.ignore_error = ignore_error

    def on_agent_start(
        self, name: str, model_name: str, input: Any, **kwargs: Any
    ) -> Any:
        cb_response = FloCallbackResponse(
            'on_agent_start', name, input=input, args=kwargs, model_name=model_name
        )
        safe_call_cb(self.func, cb_response, self.ignore_error)
        return cb_response

    def on_agent_end(
        self, name: str, model_name: str, output: Any, **kwargs: Any
    ) -> None:
        cb_response = FloCallbackResponse(
            'on_agent_end', name, output=output, args=kwargs, model_name=model_name
        )
        safe_call_cb(self.func, cb_response, self.ignore_error)
        return cb_response

    def on_agent_error(
        self,
        name: str,
        model_name: str,
        error: Union[Exception, KeyboardInterrupt],
        **kwargs: Any,
    ) -> None:
        cb_response = FloCallbackResponse(
            'on_agent_error', name, error=error, args=kwargs, model_name=model_name
        )
        safe_call_cb(self.func, cb_response, self.ignore_error)
        return cb_response


class FunctionalFloRouterCallbackImpl(FloRouterCallback):
    def __init__(self, func: Callable, ignore_error: bool = True) -> None:
        super().__init__()
        self.func = func
        self.ignore_error = ignore_error

    def on_router_start(
        self, name: str, model_name: str, input: Any, **kwargs: Any
    ) -> Any:
        cb_response = FloCallbackResponse(
            'on_router_start', name, input=input, args=kwargs, model_name=model_name
        )
        safe_call_cb(self.func, cb_response, self.ignore_error)
        return cb_response

    def on_router_end(
        self, name: str, model_name: str, output: Any, **kwargs: Any
    ) -> None:
        cb_response = FloCallbackResponse(
            'on_router_end', name, output=output, args=kwargs, model_name=model_name
        )
        safe_call_cb(self.func, cb_response, self.ignore_error)
        return cb_response

    def on_router_error(
        self,
        name: str,
        model_name: str,
        error: Union[Exception, KeyboardInterrupt],
        **kwargs: Any,
    ) -> None:
        cb_response = FloCallbackResponse(
            'on_router_error', name, error=error, args=kwargs, model_name=model_name
        )
        safe_call_cb(self.func, cb_response, self.ignore_error)
        return cb_response


class FloCallback(
    FunctionalFloToolCallbackImpl,
    FunctionalFloAgentCallbackImpl,
    FunctionalFloRouterCallbackImpl,
):
    def __init__(self, func: Callable, ignore_error: bool = True) -> None:
        FunctionalFloToolCallbackImpl.__init__(self, func, ignore_error)
        FunctionalFloAgentCallbackImpl.__init__(self, func, ignore_error)
        FunctionalFloRouterCallbackImpl.__init__(self, func, ignore_error)


def flo_tool_callback(func: Callable, ignore_error=True) -> FloToolCallback:
    return FunctionalFloToolCallbackImpl(func, ignore_error)


def flo_agent_callback(func: Callable, ignore_error=True) -> FloToolCallback:
    return FunctionalFloAgentCallbackImpl(func, ignore_error)


def flo_router_callback(func: Callable, ignore_error=True) -> FloRouterCallback:
    return FunctionalFloRouterCallbackImpl(func, ignore_error)


def flo_call_back(func: Callable, ignore_error=True) -> FloRouterCallback:
    return FloCallback(func, ignore_error)



import re
from flo_ai.error.flo_exception import FloException

name_regex = r'^[a-zA-Z0-9-_]+$'


def raise_for_name_error(string):
    if not re.match(name_regex, string):
        raise FloException(
            'Name must contain only alphanumeric characters and hyphens.'
        )



from typing import Optional
from flo_ai.state.flo_session import FloSession
from flo_ai.router.flo_supervisor import FloSupervisor
from flo_ai.router.flo_llm_router import FloLLMRouter
from flo_ai.router.flo_linear import FloLinear
from flo_ai.yaml.config import TeamConfig
from flo_ai.models.flo_team import FloTeam
from flo_ai.router.flo_router import FloRouter
from flo_ai.error.flo_exception import FloException
from flo_ai.constants.common_constants import DOCUMENTATION_ROUTER_ANCHOR


class FloRouterFactory:
    @staticmethod
    def create(
        session: FloSession, team_config: TeamConfig, flo_team: FloTeam
    ) -> FloRouter:
        router_kind = team_config.router.kind
        router_model = FloRouterFactory.__resolve_model(
            session, team_config.router.model
        )
        if router_kind == 'supervisor':
            return FloSupervisor.Builder(
                session,
                team_config.name,
                flo_team,
                llm=router_model,
                model_nick_name=team_config.router.model,
            ).build()
        elif router_kind == 'linear':
            return FloLinear.Builder(session, team_config.name, flo_team).build()
        elif router_kind == 'llm':
            return FloLLMRouter.Builder(
                session,
                team_config.router.name,
                flo_team,
                llm=router_model,
                model_nick_name=team_config.router.model,
            ).build()
        else:
            raise Exception(f"""Unknown router type: {router_kind}. 
                            The supported types are supervisor, linear and llm. 
                            Check the documentation @ {DOCUMENTATION_ROUTER_ANCHOR}""")

    @staticmethod
    def __resolve_model(session: FloSession, model_name: Optional[str] = None):
        if model_name is None:
            return session.llm
        if model_name not in session.models:
            raise FloException(
                f"""Model not found: {model_name}.
                The model you would like to use should be registered to the session using session.register_model api, 
                and the same model name should be used here instead of `{model_name}`"""
            )
        return session.models[model_name]



from typing import List, Union
from langchain_core.vectorstores import VectorStoreRetriever
from langchain.output_parsers import PydanticOutputParser
from langchain_core.prompts import PromptTemplate
from pydantic import BaseModel, Field
from flo_ai.state.flo_session import FloSession
from langchain.retrievers.multi_query import MultiQueryRetriever


class LineList(BaseModel):
    lines: List[str] = Field(description='Lines of text')


class LineListOutputParser(PydanticOutputParser):
    def __init__(self) -> None:
        super().__init__(pydantic_object=LineList)

    def parse(self, text: str) -> LineList:
        lines = text.strip().split('\n')
        return LineList(lines=lines)


class FloMultiQueryRetriever:
    def __init__(self, retriever) -> None:
        self.retriever = retriever


class FloMultiQueryRetriverBuilder:
    def __init__(
        self,
        session: FloSession,
        retriver: VectorStoreRetriever,
        query_prompt: Union[str, None] = None,
    ) -> None:
        self.session = session
        self.retriver = retriver
        self.output_parser = LineListOutputParser()

        self.prompt = PromptTemplate(
            input_variables=['question'],
            template="""You are an AI language model assistant. Your task is to generate three 
            different versions of the given user question to retrieve relevant documents from a vector 
            database. By generating multiple perspectives on the user question, your goal is to help
            the user overcome some of the limitations of the distance-based similarity search. 
            Provide these alternative questions separated by newlines.
            Original question: {question}"""
            if query_prompt is None
            else query_prompt,
        )

    def build(self):
        multi_query_retriever = MultiQueryRetriever.from_llm(
            retriever=self.retriver, llm=self.session.llm, prompt=self.prompt
        )
        return FloMultiQueryRetriever(multi_query_retriever)



from typing import Any, Dict, List, Union
from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import AgentAction, AgentFinish, LLMResult
from flo_ai.common.flo_logger import get_logger
from flo_ai.state.flo_callbacks import FloToolCallback


class FloLangchainLogger(BaseCallbackHandler):
    def __init__(self, session_id: str, tool_callbacks: List[FloToolCallback] = []):
        self.session_id = session_id
        self.tool_callbacks = tool_callbacks

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        get_logger().debug(f'onLLMStart: {prompts}', self)

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        get_logger().debug(f'onNewToken: {token}', self)

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        get_logger().debug(f'onLLMEnd: {response.generations}', self)

    def on_llm_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        get_logger().debug(f'onLLMEnd: {error}', self)

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        get_logger().debug(f'onChainStart: {inputs}', self)

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        get_logger().debug(f'onChainEnd: {outputs}', self)

    def on_chain_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        get_logger().debug(f'onChainError: {error}', self)

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> None:
        get_logger().debug(f'onToolStart: {input_str}', self)
        [
            x.on_tool_start(serialized['name'], kwargs['inputs'], kwargs)
            for x in self.tool_callbacks
        ]

    def on_tool_end(self, output: str, **kwargs: Any) -> None:
        get_logger().debug(f'onToolEnd: {output}', self)
        [x.on_tool_end(kwargs['name'], output, kwargs) for x in self.tool_callbacks]

    def on_tool_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        get_logger().debug(f'onToolError: {error}', self)
        [x.on_tool_error(kwargs['name'], error, kwargs) for x in self.tool_callbacks]

    def on_text(self, text: str, **kwargs: Any) -> None:
        get_logger().debug(f'onText: {text}', self)

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        get_logger().debug(f'onAgentAction: {action.tool} - {action.tool_input}', self)

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        get_logger().debug(f'onAgentFinish: {finish.return_values}', self)



import random
import string


def random_str(length: int = 5):
    letters = string.ascii_letters + string.digits
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str



from langchain_core.runnables import Runnable
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.state.flo_session import FloSession
from typing import Union, Optional
from langchain_core.output_parsers import StrOutputParser
from flo_ai.models.flo_executable import ExecutableType


class FloLLMAgent(ExecutableFlo):
    def __init__(self, name: str, executor: Runnable, model_name: str) -> None:
        super().__init__(name, executor, ExecutableType.llm)
        self.executor: Runnable = executor
        self.model_name: str = model_name

    @staticmethod
    def create(
        session: FloSession,
        name: str,
        job: str,
        role: Optional[str] = None,
        llm: Union[BaseLanguageModel, None] = None,
    ):
        model_name = 'default' if llm is None else llm.name
        return FloLLMAgent.Builder(
            session=session,
            name=name,
            job=job,
            role=role,
            llm=llm,
            model_name=model_name,
        ).build()

    class Builder:
        def __init__(
            self,
            session: FloSession,
            name: str,
            job: str,
            role: Optional[str] = None,
            llm: Union[BaseLanguageModel, None] = None,
            model_name: str = None,
        ) -> None:
            self.model_name = model_name
            prompt: Union[ChatPromptTemplate, str] = job

            self.name: str = name
            self.llm = llm if llm is not None else session.llm
            # TODO improve to add more context of what other agents are available
            system_prompts = (
                [('system', 'You are a {}'.format(role)), ('system', prompt)]
                if role is not None
                else [('system', prompt)]
            )
            system_prompts.append(MessagesPlaceholder(variable_name='messages'))
            self.prompt: ChatPromptTemplate = (
                ChatPromptTemplate.from_messages(system_prompts)
                if isinstance(prompt, str)
                else prompt
            )

        def build(self) -> Runnable:
            executor = self.prompt | self.llm | StrOutputParser()
            return FloLLMAgent(self.name, executor, self.model_name)



from langchain_core.runnables import Runnable
from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.state.flo_session import FloSession
from flo_ai.models.flo_executable import ExecutableType


class FloToolAgent(ExecutableFlo):
    def __init__(self, name: str, executor: Runnable, model_name: str) -> None:
        super().__init__(name, executor, ExecutableType.tool)
        self.executor: Runnable = executor
        self.model_name: str = model_name

    @staticmethod
    def create(
        session: FloSession,
        name: str,
        tool: Runnable,
    ):
        model_name = 'default'
        return FloToolAgent.Builder(
            session=session,
            name=name,
            tool_runnable=tool,
            model_name=model_name,
        ).build()

    class Builder:
        def __init__(
            self,
            session: FloSession,
            name: str,
            tool_runnable: Runnable,
            model_name: str,
        ) -> None:
            self.name: str = name
            self.runnable = tool_runnable
            self.model_name = model_name

        def build(self) -> Runnable:
            return FloToolAgent(self.name, self.runnable, self.model_name)



class FloValidationException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class FloIllegalStateException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



from flo_ai.models.flo_member import FloMember
from flo_ai.state.flo_session import FloSession


class FloTeam:
    def __init__(
        self, session: FloSession, name: str, members: list[FloMember]
    ) -> None:
        self.name = name
        self.members = members
        self.session = session

    @staticmethod
    def create(session: FloSession, name: str, members: list[FloMember]):
        return FloTeam.Builder(session=session, name=name, members=members).build()

    class Builder:
        def __init__(
            self, session: FloSession, name: str, members: list[FloMember]
        ) -> None:
            from flo_ai import Flo

            self.name = name
            self.session = session
            self.members = list(
                map(lambda x: x.runnable if isinstance(x, Flo) else x, members)
            )
            self.member_names = list(map(lambda x: x.name, self.members))

        def build(self):
            return FloTeam(name=self.name, session=self.session, members=self.members)



from typing import Optional
from langchain_core.runnables import Runnable
from flo_ai.state.flo_session import FloSession
from flo_ai.models.flo_executable import ExecutableFlo, ExecutableType
from flo_ai.models.delegate import Delegate
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.language_models import BaseLanguageModel
from pydantic import BaseModel, Field
from langchain_core.output_parsers import JsonOutputParser


class NextAgent(BaseModel):
    next: str = Field(description='Name of the next member to be called')


class FloDelegatorAgent(ExecutableFlo):
    def __init__(
        self,
        session: FloSession,
        executor: Runnable,
        delegate: Delegate,
        name: str,
        model_name: str,
    ) -> None:
        super().__init__(name, executor, ExecutableType.delegator)
        self.session = session
        self.delegate = delegate
        self.executor = executor
        self.model_name = model_name

    @staticmethod
    def create(
        session: FloSession,
        name: str,
        job: str,
        to: Delegate,
        llm: Optional[BaseLanguageModel] = None,
    ):
        model_name = 'default' if llm is None else llm.name
        return FloDelegatorAgent.Builder(
            session=session,
            name=name,
            job=job,
            delegate=to,
            llm=llm,
            model_name=model_name,
        ).build()

    class Builder:
        def __init__(
            self,
            session: FloSession,
            name: str,
            job: str,
            delegate: Delegate,
            llm: Optional[BaseLanguageModel] = None,
            model_name: str = None,
        ) -> None:
            self.session = session
            self.name = name
            self.to = delegate
            delegator_base_system_message = (
                'You are a delegator tasked with routing a conversation between the'
                ' following {member_type}: {members}. Given the following rules,'
                ' respond with the worker to act next '
            )
            self.model_name = model_name
            self.llm = session.llm if llm is None else llm
            self.options = delegate.to
            self.parser = JsonOutputParser(pydantic_object=NextAgent)
            self.llm_router_prompt = ChatPromptTemplate.from_messages(
                [
                    ('system', delegator_base_system_message),
                    MessagesPlaceholder(variable_name='messages'),
                    ('system', 'Rules: {delegator_rules}'),
                    (
                        'system',
                        'Given the conversation above, who should act next?'
                        'Select one of: {options} \n {format_instructions}',
                    ),
                ]
            ).partial(
                options=str(self.options),
                members=', '.join(self.options),
                member_type='agents',
                delegator_rules=job,
                format_instructions=self.parser.get_format_instructions(),
            )

        def build(self):
            chain = self.llm_router_prompt | self.llm | self.parser

            return FloDelegatorAgent(
                session=self.session,
                name=self.name,
                delegate=self.to,
                executor=chain,
                model_name=self.model_name,
            )



from flo_ai.models.flo_member import FloMember
from langchain_core.runnables import Runnable
from langchain_core.messages import HumanMessage
from enum import Enum
from flo_ai.state.flo_state import STATE_NAME_MESSAGES


class ExecutableType(Enum):
    agentic = 'agentic'
    llm = 'llm'
    tool = 'tool'
    reflection = 'reflection'
    delegator = 'delegator'
    team = 'team'
    router = 'router'

    @staticmethod
    def isAgent(type: 'ExecutableType'):
        match type:
            case ExecutableType.agentic:
                return True
            case ExecutableType.llm:
                return True
            case ExecutableType.tool:
                return True
        return False


class ExecutableFlo(FloMember):
    def __init__(
        self, name: str, runnable: Runnable, type: str = ExecutableType.team
    ) -> None:
        super().__init__(name, type)
        self.runnable = runnable

    def stream(self, work, config=None):
        return self.runnable.stream(
            {STATE_NAME_MESSAGES: [HumanMessage(content=work)]}, config
        )

    def astream(self, work, config=None):
        return self.runnable.astream(
            {STATE_NAME_MESSAGES: [HumanMessage(content=work)]}, config
        )

    def invoke(self, work, config=None):
        return self.runnable.invoke(
            {
                STATE_NAME_MESSAGES: [HumanMessage(content=work)],
            },
            config,
        )

    def ainvoke(self, work, config=None):
        return self.runnable.ainvoke(
            {
                STATE_NAME_MESSAGES: [HumanMessage(content=work)],
            },
            config,
        )

    def draw(self, xray=True):
        return self.runnable.get_graph().draw_mermaid_png()



import functools
from flo_ai.models.flo_agent import FloAgent
from flo_ai.models.flo_reflection_agent import FloReflectionAgent
from flo_ai.models.flo_routed_team import FloRoutedTeam
from flo_ai.models.delegate import Delegate
from langchain.agents import AgentExecutor
from flo_ai.state.flo_state import TeamFloAgentState, STATE_NAME_MESSAGES
from langchain_core.messages import HumanMessage
from flo_ai.models.flo_executable import ExecutableType
from flo_ai.state.flo_session import FloSession
from typing import Optional, Type, List
from flo_ai.state.flo_callbacks import FloAgentCallback, FloRouterCallback, FloCallback


class FloNode:
    def __init__(
        self,
        func: functools.partial,
        name: str,
        kind: ExecutableType,
        delegate: Optional[Delegate] = None,
    ) -> None:
        self.name = name
        self.func = func
        self.kind: ExecutableType = kind
        self.delegate = delegate

    class Builder:
        def __init__(self, session: FloSession) -> None:
            self.session = session

        def build_from_agent(self, flo_agent: FloAgent) -> 'FloNode':
            agent_func = functools.partial(
                FloNode.Builder.__teamflo_agent_node,
                agent=flo_agent.runnable,
                name=flo_agent.name,
                session=self.session,
                model_name=flo_agent.model_name,
            )
            return FloNode(agent_func, flo_agent.name, flo_agent.type)

        def build_from_reflection(self, flo_agent: FloReflectionAgent) -> 'FloNode':
            agent_func = functools.partial(
                FloNode.Builder.__teamflo_agent_node,
                agent=flo_agent.runnable,
                name=flo_agent.name,
                session=self.session,
                model_name=flo_agent.model_name,
            )
            return FloNode(
                agent_func, flo_agent.name, flo_agent.type, delegate=flo_agent.delegate
            )

        def build_from_team(self, flo_team: FloRoutedTeam) -> 'FloNode':
            team_chain = (
                functools.partial(
                    FloNode.Builder.__teamflo_team_node, members=flo_team.runnable.nodes
                )
                | flo_team.runnable
            )
            return FloNode(
                (
                    FloNode.Builder.__get_last_message
                    | team_chain
                    | FloNode.Builder.__join_graph
                ),
                flo_team.name,
                flo_team.type,
            )

        def build_from_router(self, flo_router) -> 'FloNode':
            router_func = functools.partial(
                FloNode.Builder.__teamflo_router_node,
                agent=flo_router.executor,
                name=flo_router.name,
                session=self.session,
                model_name=flo_router.model_name,
            )
            return FloNode(router_func, flo_router.name, flo_router.type)

        def build_from_delegator(self, flo_router) -> 'FloNode':
            router_func = functools.partial(
                FloNode.Builder.__teamflo_router_node,
                agent=flo_router.executor,
                name=flo_router.name,
                session=self.session,
                model_name=flo_router.model_name,
            )
            return FloNode(
                router_func,
                flo_router.name,
                flo_router.type,
                delegate=flo_router.delegate,
            )

        @staticmethod
        def __teamflo_agent_node(
            state: TeamFloAgentState,
            agent: AgentExecutor,
            name: str,
            session: FloSession,
            model_name: str,
        ):
            agent_cbs: List[FloAgentCallback] = FloNode.Builder.__filter_callbacks(
                session, FloAgentCallback
            )
            flo_cbs: List[FloCallback] = FloNode.Builder.__filter_callbacks(
                session, FloCallback
            )
            [
                callback.on_agent_start(name, model_name, state['messages'], **{})
                for callback in agent_cbs
            ]
            [
                callback.on_agent_start(name, model_name, state['messages'], **{})
                for callback in flo_cbs
            ]
            try:
                result = agent.invoke(state)
                output = result if isinstance(result, str) else result['output']
            except Exception as e:
                [
                    callback.on_agent_error(name, model_name, e, **{})
                    for callback in agent_cbs
                ]
                [
                    callback.on_agent_error(name, model_name, e, **{})
                    for callback in flo_cbs
                ]
                raise e
            [
                callback.on_agent_end(name, model_name, output, **{})
                for callback in agent_cbs
            ]
            [
                callback.on_agent_start(name, model_name, output, **{})
                for callback in flo_cbs
            ]
            return {STATE_NAME_MESSAGES: [HumanMessage(content=output, name=name)]}

        @staticmethod
        def __filter_callbacks(session: FloSession, type: Type):
            cbs = session.callbacks
            return list(filter(lambda callback: isinstance(callback, type), cbs))

        @staticmethod
        def __teamflo_router_node(
            state: TeamFloAgentState,
            agent: AgentExecutor,
            name: str,
            session: FloSession,
            model_name: str,
        ):
            agent_cbs: List[FloRouterCallback] = FloNode.Builder.__filter_callbacks(
                session, FloRouterCallback
            )
            flo_cbs: List[FloCallback] = FloNode.Builder.__filter_callbacks(
                session, FloCallback
            )
            [
                callback.on_router_start(name, model_name, state['messages'], **{})
                for callback in agent_cbs
            ]
            [
                callback.on_router_start(name, model_name, state['messages'], **{})
                for callback in flo_cbs
            ]
            try:
                result = agent.invoke(state)
                nextNode = result if isinstance(result, str) else result['next']
            except Exception as e:
                [
                    callback.on_router_error(name, model_name, e, **{})
                    for callback in agent_cbs
                ]
                [
                    callback.on_router_error(name, model_name, e, **{})
                    for callback in flo_cbs
                ]
                raise e
            [
                callback.on_router_end(name, model_name, nextNode, **{})
                for callback in agent_cbs
            ]
            [
                callback.on_router_start(name, model_name, nextNode, **{})
                for callback in flo_cbs
            ]
            return {'next': nextNode}

        @staticmethod
        def __get_last_message(state: TeamFloAgentState) -> str:
            return state[STATE_NAME_MESSAGES][-1].content

        @staticmethod
        def __join_graph(response: dict):
            return {STATE_NAME_MESSAGES: [response[STATE_NAME_MESSAGES][-1]]}

        @staticmethod
        def __teamflo_team_node(message: str, members: list[str]):
            results = {
                STATE_NAME_MESSAGES: [HumanMessage(content=message)],
                'team_members': ', '.join(members),
            }
            return results



from langchain_core.vectorstores import VectorStoreRetriever
from langchain_core.runnables import RunnableParallel, Runnable
from flo_ai.state.flo_session import FloSession
from langchain.schema.output_parser import StrOutputParser
from langchain.schema.runnable import RunnablePassthrough
from flo_ai.retrievers.flo_multi_query import FloMultiQueryRetriverBuilder
from langchain.retrievers import ContextualCompressionRetriever
from langchain.retrievers.document_compressors import DocumentCompressorPipeline
from flo_ai.retrievers.flo_compression_pipeline import FloCompressionPipeline
from functools import partial
from pydantic import BaseModel, Field
from langchain_core.tools import Tool
from typing import Optional
from langchain_core.callbacks import Callbacks
from langchain_core.prompts import (
    BasePromptTemplate,
    PromptTemplate,
    aformat_document,
    format_document,
    ChatPromptTemplate,
    MessagesPlaceholder,
)
from typing import List


class FloRagBaseMessage(BaseModel):
    content: str


class FloRagToolInput(BaseModel):
    messages: List[FloRagBaseMessage] = Field(
        description='query to look up in the vector store'
    )


def _get_relevant_documents(
    messages: List[FloRagBaseMessage],
    retriever: VectorStoreRetriever,
    document_prompt: BasePromptTemplate,
    document_separator: str,
    callbacks: Callbacks = None,
) -> str:
    docs = retriever.invoke(messages[-1].content, config={'callbacks': callbacks})
    return document_separator.join(
        format_document(doc, document_prompt) for doc in docs
    )


async def _aget_relevant_documents(
    messages: List[str],
    retriever: VectorStoreRetriever,
    document_prompt: BasePromptTemplate,
    document_separator: str,
    callbacks: Callbacks = None,
) -> str:
    docs = await retriever.ainvoke(
        messages[-1].content, config={'callbacks': callbacks}
    )
    return document_separator.join(
        [await aformat_document(doc, document_prompt) for doc in docs]
    )


class FloRagBuilder:
    def __init__(self, session: FloSession, retriever: VectorStoreRetriever) -> None:
        self.session = session
        self.retriever = retriever
        self.default_prompt = ChatPromptTemplate.from_messages(
            [
                (
                    'system',
                    """You are an assistant for question-answering tasks. 
                 Use the following pieces of retrieved context to answer the question. 
                 If you don't know the answer, just say that you don't know. 
                 Use three sentences maximum and keep the answer concise.

                 Here is the context:
                 {context}
                 
                 """,
                ),
                MessagesPlaceholder(variable_name='chat_history'),
                ('human', '{question}'),
            ]
        )
        self.history_aware_retriever = self.__create_history_aware_retriever()

    def with_prompt(self, prompt: ChatPromptTemplate):
        self.default_prompt = prompt
        return self

    def with_multi_query(self, prompt=None):
        builder = FloMultiQueryRetriverBuilder(
            session=self.session, retriver=self.retriever, query_prompt=prompt
        )
        multi_query_retriever = builder.build()
        self.retriever = multi_query_retriever.retriever
        return self

    def with_compression(self, pipeline: FloCompressionPipeline):
        pipeline_compressor = DocumentCompressorPipeline(transformers=pipeline.get())
        compression_retriever = ContextualCompressionRetriever(
            base_compressor=pipeline_compressor, base_retriever=self.retriever
        )
        self.retriever = compression_retriever
        return self

    def __create_history_aware_retriever(self):
        contextualize_q_system_prompt = """Given a chat history and the latest user question \
        which might reference context in the chat history, formulate a standalone question \
        which can be understood without the chat history. Do NOT answer the question, \
        just reformulate it if needed and otherwise return it as is."""

        contextualize_q_prompt = ChatPromptTemplate.from_messages(
            [
                ('system', contextualize_q_system_prompt),
                MessagesPlaceholder('chat_history'),
                ('human', '{question}'),
            ]
        )
        self.history_aware_retriever = (
            contextualize_q_prompt | self.session.llm | StrOutputParser()
        )
        return self.history_aware_retriever

    def __get_retriever(self):
        def __precontext_retriver(input_prompt: dict):
            if input_prompt.get('chat_history'):
                return self.history_aware_retriever
            else:
                return input_prompt['question']

        return __precontext_retriver | self.retriever

    def __format_docs(self, docs):
        return '\n\n'.join(doc.page_content for doc in docs)

    def __get_optional_chat_history(self, x):
        return x['chat_history'] if 'chat_history' in x else []

    def __build_history_aware_rag(self):
        rag_chain = (
            RunnablePassthrough.assign(
                context=(lambda x: x['context']),
            )
            | self.default_prompt
            | self.session.llm
        )

        rag_chain_with_source = RunnableParallel(
            {
                'context': self.__get_retriever() | self.__format_docs,
                'question': RunnablePassthrough(),
                'chat_history': lambda x: self.__get_optional_chat_history(x),
            }
        ).assign(answer=rag_chain)
        return rag_chain_with_source

    def build_rag(self):
        return self.__build_history_aware_rag()

    def build_retriever_tool(self, name, description):
        return self.__create_retriever_tool(self.retriever, name, description)

    @staticmethod
    def __get_rag_answer(messages: List[FloRagBaseMessage], runnable: Runnable):
        question = messages[-1].content
        chat_history = messages[:-1]
        result = runnable.invoke({'question': question, 'chat_history': chat_history})
        return result['answer'].content

    @staticmethod
    async def __aget_rag_answer(messages: List[FloRagBaseMessage], runnable: Runnable):
        question = messages[-1].content
        chat_history = messages[:-1]
        result = await runnable.ainvoke(
            {'question': question, 'chat_history': chat_history}
        )
        return result['answer'].content

    def __create_retriever_tool(
        self,
        retriever: VectorStoreRetriever,
        name: str,
        description: str,
        *,
        document_prompt: Optional[BasePromptTemplate] = None,
        document_separator: str = '\n',
    ) -> Tool:
        document_prompt = document_prompt or PromptTemplate.from_template(
            '{page_content}'
        )
        func = partial(
            _get_relevant_documents,
            retriever=retriever,
            document_prompt=document_prompt,
            document_separator=document_separator,
        )
        afunc = partial(
            _aget_relevant_documents,
            retriever=retriever,
            document_prompt=document_prompt,
            document_separator=document_separator,
        )
        return Tool(
            name=name,
            description=description,
            func=func,
            coroutine=afunc,
            args_schema=FloRagToolInput,
        )

    @staticmethod
    def __create_rag_tool(runnable_rag: Runnable, name: str, description: str) -> Tool:
        func = partial(FloRagBuilder.__get_rag_answer, runnable=runnable_rag)

        afunc = partial(FloRagBuilder.__aget_rag_answer, runnable=runnable_rag)

        return Tool(
            name=name,
            description=description,
            func=func,
            coroutine=afunc,
            args_schema=FloRagToolInput,
        )

    def build_rag_tool(self, name, description) -> Tool:
        rag = self.__build_history_aware_rag()
        return FloRagBuilder.__create_rag_tool(rag, name, description)



from langchain_core.tools import BaseTool
from langchain.agents import AgentExecutor
from langchain.agents import create_tool_calling_agent
from langchain_core.runnables import Runnable
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.state.flo_session import FloSession
from typing import Union, Optional, Callable
from flo_ai.models.flo_executable import ExecutableType


class FloAgent(ExecutableFlo):
    def __init__(
        self,
        name: str,
        agent: Runnable,
        executor: AgentExecutor,
        model_name: str,
    ) -> None:
        super().__init__(name, executor, ExecutableType.agentic)
        self.model_name = model_name
        self.agent: Runnable = (agent,)
        self.executor: AgentExecutor = executor

    @staticmethod
    def create(
        session: FloSession,
        name: str,
        job: str,
        tools: list[BaseTool],
        role: Optional[str] = None,
        on_error: Union[str, Callable] = True,
        llm: Union[BaseLanguageModel, None] = None,
    ):
        model_name = 'default' if llm is None else llm.name
        return FloAgent.Builder(
            session=session,
            name=name,
            job=job,
            tools=tools,
            role=role,
            on_error=on_error,
            llm=llm,
            model_name=model_name,
        ).build()

    class Builder:
        def __init__(
            self,
            session: FloSession,
            name: str,
            job: str,
            tools: list[BaseTool],
            role: Optional[str] = None,
            verbose: bool = False,
            llm: Union[BaseLanguageModel, None] = None,
            on_error: Union[str, Callable] = True,
            model_name: Union[str, None] = 'default',
        ) -> None:
            prompt: Union[ChatPromptTemplate, str] = job
            self.name: str = name
            self.model_name = model_name
            self.llm = llm if llm is not None else session.llm
            system_prompts = (
                [('system', 'You are a {}'.format(role)), ('system', prompt)]
                if role is not None
                else [('system', prompt)]
            )
            system_prompts.append(MessagesPlaceholder(variable_name='messages'))
            system_prompts.append(MessagesPlaceholder(variable_name='agent_scratchpad'))
            self.prompt: ChatPromptTemplate = (
                ChatPromptTemplate.from_messages(system_prompts)
                if isinstance(prompt, str)
                else prompt
            )
            self.tools: list[BaseTool] = tools
            self.verbose = verbose
            self.on_error = on_error

        def build(self) -> AgentExecutor:
            agent = create_tool_calling_agent(self.llm, self.tools, self.prompt)
            executor = AgentExecutor(
                agent=agent,
                tools=self.tools,
                verbose=self.verbose,
                return_intermediate_steps=True,
                handle_parsing_errors=self.on_error,
            )
            return FloAgent(self.name, agent, executor, model_name=self.model_name)



from flo_ai.core import Flo as Flo
from flo_ai.tools.flo_tool import flotool as flotool
from flo_ai.models.flo_team import FloTeam as FloTeam
from flo_ai.models.flo_agent import FloAgent as FloAgent
from flo_ai.router.flo_linear import FloLinear as FloLinear
from flo_ai.router.flo_router import FloRouter as FloRouter
from flo_ai.state.flo_session import FloSession as FloSession
from flo_ai.models.flo_llm_agent import FloLLMAgent as FloLLMAgent
from flo_ai.models.flo_tool_agent import FloToolAgent as FloToolAgent
from flo_ai.router.flo_llm_router import FloLLMRouter as FloLLMRouter
from flo_ai.router.flo_supervisor import FloSupervisor as FloSupervisor
from flo_ai.retrievers.flo_retriever import FloRagBuilder as FloRagBuilder
from flo_ai.models.flo_delegation_agent import FloDelegatorAgent as FloDelegatorAgent
from flo_ai.models.flo_reflection_agent import FloReflectionAgent as FloReflectionAgent



from flo_ai.router.flo_router import FloRouter
from langgraph.graph import StateGraph, END, START
from flo_ai.state.flo_state import TeamFloAgentState
from flo_ai.models.flo_routed_team import FloRoutedTeam
from flo_ai.models.flo_team import FloTeam
from flo_ai.state.flo_session import FloSession
from flo_ai.models.flo_executable import ExecutableType


class FloLinear(FloRouter):
    def __init__(self, session: FloSession, name: str, flo_team: FloTeam):
        super().__init__(
            session=session,
            name=name,
            flo_team=flo_team,
            executor=None,
            model_name=None,
        )

    def build_graph(self):
        flo_agent_nodes = [self.build_node(member) for member in self.members]
        workflow = StateGraph(TeamFloAgentState)

        for flo_node in flo_agent_nodes:
            agent_name = flo_node.name
            workflow.add_node(agent_name, flo_node.func)

        start_node = flo_agent_nodes[0]
        end_node = flo_agent_nodes[-1]
        workflow.add_edge(START, start_node.name)
        for i in range(len(flo_agent_nodes) - 1):
            parent_node = flo_agent_nodes[i]
            child_node = flo_agent_nodes[i + 1]
            next_node = (
                flo_agent_nodes[i + 2] if (i + 2) < len(flo_agent_nodes) else END
            )
            if parent_node.kind == ExecutableType.reflection:
                self.add_reflection_edge(workflow, parent_node, child_node)
                continue
            if child_node.kind == ExecutableType.delegator:
                self.add_delegation_edge(workflow, parent_node, child_node, next_node)
                continue

            if (
                child_node.kind != ExecutableType.reflection
                and parent_node.kind != ExecutableType.delegator
            ):
                workflow.add_edge(parent_node.name, child_node.name)

        if end_node.kind == ExecutableType.reflection:
            self.add_reflection_edge(workflow, end_node, END)
        elif end_node.kind != ExecutableType.delegator:
            workflow.add_edge(end_node.name, END)

        workflow_graph = workflow.compile()

        return FloRoutedTeam(self.flo_team.name, workflow_graph)

    @staticmethod
    def create(session: FloSession, name: str, team: FloTeam):
        return FloLinear.Builder(session=session, name=name, flo_team=team).build()

    class Builder:
        def __init__(self, session: FloSession, name: str, flo_team: FloTeam) -> None:
            self.name = name
            self.session = session
            self.team = flo_team

        def build(self):
            return FloLinear(self.session, self.name, self.team)
